package com.major.retrofit.interceptor;

import com.major.retrofit.util.LogUtil;
import com.major.retrofit.util.ZipHelper;

import java.io.EOFException;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.UnsupportedCharsetException;
import java.util.concurrent.TimeUnit;

import okhttp3.Connection;
import okhttp3.Headers;
import okhttp3.Interceptor;
import okhttp3.MediaType;
import okhttp3.Protocol;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.ResponseBody;
import okhttp3.internal.http.HttpHeaders;
import okio.Buffer;
import okio.BufferedSource;

/**
 * @desc: 参考 {@link okhttp3.logging.HttpLoggingInterceptor}
 * @author: Major
 * @since: 2017/12/17 0:02
 */
public class HttpLoggingInterceptor implements Interceptor{

    private static final Charset UTF8 = Charset.forName("UTF-8");

    public enum Level{
        /** No logs. */
        NONE,
        /**
         * Logs request and response lines.
         * <p>
         * <p>Example:
         * <pre>{@code
         * --> POST /greeting http/1.1 (3-byte body)
         *
         * <-- 200 OK (22ms, 6-byte body)
         * }</pre>
         */
        BASIC,
        /**
         * Logs request and response lines and their respective headers.
         * <p>
         * <p>Example:
         * <pre>{@code
         * --> POST /greeting http/1.1
         * Host: example.com
         * Content-Type: plain/text
         * Content-Length: 3
         * --> END POST
         *
         * <-- 200 OK (22ms)
         * Content-Type: plain/text
         * Content-Length: 6
         * <-- END HTTP
         * }</pre>
         */
        HEADERS,
        /**
         * Logs request and response lines and their respective headers and bodies (if present).
         * <p>
         * <p>Example:
         * <pre>{@code
         * --> POST /greeting http/1.1
         * Host: example.com
         * Content-Type: plain/text
         * Content-Length: 3
         *
         * Hi?
         * --> END POST
         *
         * <-- 200 OK (22ms)
         * Content-Type: plain/text
         * Content-Length: 6
         *
         * Hello!
         * <-- END HTTP
         * }</pre>
         */
        BODY
    }

    public interface Logger{

        void log(String message);

        /** A {@link HttpLoggingInterceptor.Logger} defaults output appropriate for the current platform. */
        HttpLoggingInterceptor.Logger DEFAULT = new HttpLoggingInterceptor.Logger(){
            @Override
            public void log(String message){
//                Platform.get().log(INFO, message, null);
                LogUtil.i(message);
            }
        };
    }

    public HttpLoggingInterceptor(){
        this(HttpLoggingInterceptor.Logger.DEFAULT);
    }

    public HttpLoggingInterceptor(HttpLoggingInterceptor.Logger logger){
        this.logger = logger;
    }

    private final HttpLoggingInterceptor.Logger logger;

    private volatile HttpLoggingInterceptor.Level level = HttpLoggingInterceptor.Level.NONE;

    /** Change the level at which this interceptor logs. */
    public HttpLoggingInterceptor setLevel(HttpLoggingInterceptor.Level level){
        if(level == null){
            throw new NullPointerException("level == null. Use Level.NONE instead.");
        }
        this.level = level;
        return this;
    }

    public HttpLoggingInterceptor.Level getLevel(){
        return level;
    }

    @Override
    public Response intercept(Chain chain) throws IOException{
        HttpLoggingInterceptor.Level level = this.level;

        Request request = chain.request();
        if(level == HttpLoggingInterceptor.Level.NONE){
            return chain.proceed(request);
        }

        boolean logBody = level == HttpLoggingInterceptor.Level.BODY;
        boolean logHeaders = logBody || level == HttpLoggingInterceptor.Level.HEADERS;

        RequestBody requestBody = request.body();
        boolean hasRequestBody = requestBody != null;

        Connection connection = chain.connection();
        Protocol protocol = connection != null ? connection.protocol() : Protocol.HTTP_1_1;
        String requestStartMessage = "--> " + request.method() + ' ' + request.url() + ' ' + protocol;
        if(!logHeaders && hasRequestBody){
            requestStartMessage += " (" + requestBody.contentLength() + "-byte body)";
        }
        logger.log(requestStartMessage);

        if(logHeaders){
            if(hasRequestBody){
                // Request body headers are only present when installed as a network interceptor. Force
                // them to be included (when available) so there values are known.
                if(requestBody.contentType() != null){
                    logger.log("Content-Type: " + requestBody.contentType());
                }
                if(requestBody.contentLength() != -1){
                    logger.log("Content-Length: " + requestBody.contentLength());
                }
            }

            Headers headers = request.headers();
            for(int i = 0, count = headers.size(); i < count; i++){
                String name = headers.name(i);
                // Skip headers from the request body as they are explicitly logged above.
                if(!"Content-Type".equalsIgnoreCase(name) && !"Content-Length".equalsIgnoreCase(name)){
                    logger.log(name + ": " + headers.value(i));
                }
            }

            if(!logBody || !hasRequestBody){
                logger.log("--> END " + request.method());
            } else if(bodyEncoded(request.headers())){
                logger.log("--> END " + request.method() + " (encoded body omitted)");
            } else {
                Buffer buffer = new Buffer();
                requestBody.writeTo(buffer);

                Charset charset = UTF8;
                MediaType contentType = requestBody.contentType();
                if(contentType != null){
                    charset = contentType.charset(UTF8);
                }

                logger.log("");
                if(isPlaintext(buffer)){
                    logger.log(buffer.readString(charset));
                    logger.log("--> END " + request.method()
                               + " (" + requestBody.contentLength() + "-byte body)");
                } else {
                    logger.log("--> END " + request.method() + " (binary "
                               + requestBody.contentLength() + "-byte body omitted)");
                }
            }
        }

        long startNs = System.nanoTime();
        Response response;
        try{
            response = chain.proceed(request);
        } catch(Exception e){
            logger.log("<-- HTTP FAILED: " + e);
            throw e;
        }
        long tookMs = TimeUnit.NANOSECONDS.toMillis(System.nanoTime() - startNs);

        ResponseBody responseBody = response.body();
        long contentLength = responseBody.contentLength();
        String bodySize = contentLength != -1 ? contentLength + "-byte" : "unknown-length";
        logger.log("<-- " + response.code() + ' ' + response.message() + ' '
                   + response.request().url() + " (" + tookMs + "ms" + (!logHeaders ? ", "
                                                                                      + bodySize + " body" : "") + ')');

        if(logHeaders){
            Headers headers = response.headers();
            for(int i = 0, count = headers.size(); i < count; i++){
                logger.log(headers.name(i) + ": " + headers.value(i));
            }

            if(!logBody || !HttpHeaders.hasBody(response)){
                logger.log("<-- END HTTP");
            } else if(bodyEncoded(response.headers())){
                String encoding = response.headers().get("Content-Encoding");
                BufferedSource source = responseBody.source();
                source.request(Long.MAX_VALUE); // Buffer the entire body.
                Buffer buffer = source.buffer();

                if(contentLength != 0){
                    logger.log("");
                    // 获取返回 json
                    String bodyString = parseContent(responseBody, encoding, buffer.clone()); // 需要传递 Buffer 的 clone
                    logger.log("bodyString " + bodyString);
                }
                logger.log("<-- END HTTP (encoded body omitted)");

            } else {
                BufferedSource source = responseBody.source();
                source.request(Long.MAX_VALUE); // Buffer the entire body.
                Buffer buffer = source.buffer();

                Charset charset = UTF8;
                MediaType contentType = responseBody.contentType();
                if(contentType != null){
                    try{
                        charset = contentType.charset(UTF8);
                    } catch(UnsupportedCharsetException e){
                        logger.log("");
                        logger.log("Couldn't decode the response body; charset is likely malformed.");
                        logger.log("<-- END HTTP");

                        return response;
                    }
                }

                if(!isPlaintext(buffer)){
                    logger.log("");
                    logger.log("<-- END HTTP (binary " + buffer.size() + "-byte body omitted)");
                    return response;
                }

                if(contentLength != 0){
                    logger.log("");
                    logger.log(buffer.clone().readString(charset));
                }

                logger.log("<-- END HTTP (" + buffer.size() + "-byte body)");
            }
        }

        return response;
    }

    /**
     * Returns true if the body in question probably contains human readable text. Uses a small sample
     * of code points to detect unicode control characters commonly used in binary file signatures.
     */
    static boolean isPlaintext(Buffer buffer){
        try{
            Buffer prefix = new Buffer();
            long byteCount = buffer.size() < 64 ? buffer.size() : 64;
            buffer.copyTo(prefix, 0, byteCount);
            for(int i = 0; i < 16; i++){
                if(prefix.exhausted()){
                    break;
                }
                int codePoint = prefix.readUtf8CodePoint();
                if(Character.isISOControl(codePoint) && !Character.isWhitespace(codePoint)){
                    return false;
                }
            }
            return true;
        } catch(EOFException e){
            return false; // Truncated UTF-8 sequence.
        }
    }

    private boolean bodyEncoded(Headers headers){
        String contentEncoding = headers.get("Content-Encoding");
        return contentEncoding != null && !contentEncoding.equalsIgnoreCase("identity");
    }

    private String parseContent(ResponseBody responseBody, String encoding, Buffer clone) {
        Charset charset = Charset.forName("UTF-8");
        MediaType contentType = responseBody.contentType();
        if (contentType != null) {
            charset = contentType.charset(charset);
        }
        if (encoding != null && encoding.equalsIgnoreCase("gzip")) {//content使用gzip压缩
            return ZipHelper.decompressForGzip(clone.readByteArray(), convertCharset(charset));//解压
        } else if (encoding != null && encoding.equalsIgnoreCase("zlib")) {//content使用zlib压缩
            return ZipHelper.decompressToStringForZlib(clone.readByteArray(), convertCharset(charset));//解压
        } else {//content没有被压缩
            return clone.readString(charset);
        }
    }

    public static String convertCharset(Charset charset) {
        String s = charset.toString();
        int i = s.indexOf("[");
        if (i == -1)
            return s;
        return s.substring(i + 1, s.length() - 1);
    }

}
